@@ -2,7 +2,7 @@ import asyncio
 import dataclasses
 import json
 import re
-from typing import Dict, Optional, Sequence, Tuple, Union
+from typing import Dict, List, Optional, Sequence, Tuple, Union
 
 import httpstan.models
 import httpstan.schemas
@@ -27,6 +27,12 @@ class DataJSONEncoder(json.JSONEncoder):
         return json.JSONEncoder.default(self, obj)
 
 
+# superficial frozendict implementation. Only used for function signatures
+class frozendict(dict):
+    def __setitem__(self, key, value):
+        raise TypeError("'frozendict' object is immutable.")
+
+
 @dataclasses.dataclass(frozen=True)
 class Model:
     """Stores data associated with and proxies calls to a Stan model.
@@ -47,7 +53,7 @@ class Model:
         if self.model_name != httpstan.models.calculate_model_name(self.program_code):
             raise ValueError("`model_name` does not match `program_code`.")
 
-    def sample(self, **kwargs) -> stan.fit.Fit:
+    def sample(self, *, num_chains=4, **kwargs) -> stan.fit.Fit:
         """Draw samples from the model.
 
         Parameters in ``kwargs`` will be passed to the default sample function.
@@ -58,16 +64,26 @@ class Model:
 
         There is one exception:  `num_chains`. `num_chains` is a
         PyStan-specific keyword argument. It indicates the number of
-        independent processes to use when drawing samples.  The default value
-        is 4.
+        independent processes to use when drawing samples.
+
 
         Returns:
             Fit: instance of Fit allowing access to draws.
 
-        """
-        return self.hmc_nuts_diag_e_adapt(**kwargs)
+        Examples:
+            User-defined initial values for parameters must be provided
+            for each chain. Typically they will be the same for each chain.
+            The following example shows how user-defined initial parameters
+            are provided:
 
-    def hmc_nuts_diag_e_adapt(self, **kwargs) -> stan.fit.Fit:
+            >>> program_code = "parameters {real y;} model {y ~ normal(0,1);}"
+            >>> posterior = stan.build(program_code)
+            >>> fit = posterior.sample(num_chains=2, init=[{"y": 3}, {"y": 3}])
+
+        """
+        return self.hmc_nuts_diag_e_adapt(num_chains=num_chains, **kwargs)
+
+    def hmc_nuts_diag_e_adapt(self, *, num_chains=4, **kwargs) -> stan.fit.Fit:
         """Draw samples from the model using ``stan::services::sample::hmc_nuts_diag_e_adapt``.
 
         Parameters in ``kwargs`` will be passed to the (Python wrapper of)
@@ -77,17 +93,17 @@ class Model:
 
         There is one exception:  `num_chains`. `num_chains` is a
         PyStan-specific keyword argument. It indicates the number of
-        independent processes to use when drawing samples.  The default value
-        is 4.
+        independent processes to use when drawing samples.
 
         Returns:
             Fit: instance of Fit allowing access to draws.
 
         """
+        kwargs["num_chains"] = num_chains
         kwargs["function"] = "stan::services::sample::hmc_nuts_diag_e_adapt"
         return self._create_fit(kwargs)
 
-    def fixed_param(self, **kwargs) -> stan.fit.Fit:
+    def fixed_param(self, *, num_chains=4, **kwargs) -> stan.fit.Fit:
         """Draw samples from the model using ``stan::services::sample::fixed_param``.
 
         Parameters in ``kwargs`` will be passed to the (Python wrapper of)
@@ -97,13 +113,13 @@ class Model:
 
         There is one exception:  `num_chains`. `num_chains` is a
         PyStan-specific keyword argument. It indicates the number of
-        independent processes to use when drawing samples.  The default value
-        is 4.
+        independent processes to use when drawing samples.
 
         Returns:
             Fit: instance of Fit allowing access to draws.
 
         """
+        kwargs["num_chains"] = num_chains
         kwargs["function"] = "stan::services::sample::fixed_param"
         return self._create_fit(kwargs)
 
@@ -124,8 +140,11 @@ class Model:
         assert "random_seed" not in payload, "`random_seed` is set in `build`."
         assert "function" in payload
 
-        num_chains = payload.pop("num_chains", 4)
+        # copy kwargs and verify everything is JSON-encodable
+        payload = json.loads(DataJSONEncoder().encode(payload))
+        num_chains = payload.pop("num_chains")
 
+        init: List[Data]
         init = payload.pop("init", [dict() for _ in range(num_chains)])
         if len(init) != num_chains:
             raise ValueError("Initial values must be provided for each chain.")
@@ -150,7 +169,7 @@ class Model:
                 "save_warmup",
                 arguments.lookup_default(arguments.Method["SAMPLE"], "save_warmup"),
             )
-            payloads.append(payload)
+            payloads.append(payload.copy())
 
         async def go():
             io = ConsoleIO()
@@ -265,7 +284,7 @@ class Model:
 
             for entry_point in stan.plugins.get_plugins():
                 Plugin = entry_point.load()
-                fit = Plugin().on_post_fit(fit)
+                fit = Plugin().on_post_sample(fit)
             return fit
 
         try:
@@ -393,14 +412,15 @@ class Model:
         return asyncio.run(go())
 
 
-def build(program_code: str, data: Optional[Data] = None, random_seed: Optional[int] = None) -> Model:
+def build(program_code: str, data: Data = frozendict(), random_seed: Optional[int] = None) -> Model:
     """Build (compile) a Stan program.
 
     Arguments:
         program_code: Stan program code describing a Stan model.
         data: A Python dictionary or mapping providing the data for the
             model. Variable names are the keys and the values are their
-            associated values. Default is an empty dictionary.
+            associated values. Default is an empty dictionary, suitable
+            for Stan programs with no `data` block.
         random_seed: Random seed, a positive integer for random number
             generation. Used to make sure that results can be reproduced.
 
@@ -413,7 +433,7 @@ def build(program_code: str, data: Optional[Data] = None, random_seed: Optional[
 
     """
     # `data` must be JSON-serializable in order to send to httpstan
-    data = json.loads(DataJSONEncoder().encode(data)) if data else {}
+    data = json.loads(DataJSONEncoder().encode(data))
 
     async def go():
         io = ConsoleIO()
